"""
The ``PrintJob`` class is a specialization of the
:class:`~asynciojobs.job.AbstractJob` class,
mostly useful for debugging, tests and tutorials.
"""

import asyncio

from .job import AbstractJob


class PrintJob(AbstractJob):
    """
    A job that just prints messages, and optionnally sleeps for some time.

    Parameters:

      messages: passed to ``print`` as-is
      sleep: optional, an int or float describing in seconds
        how long to sleep after the messages get printed
      banner: optional, a fixed text printed out before the messages
       like e.g. ``40*'='``; it won't make it into ``details()``
      scheduler: passed to :class:``AbstractJob``
      required: passed to :class:``AbstractJob``
      label: passed to :class:``AbstractJob``
    """

    def __init__(self, *messages, sleep=None, banner=None,
                 # these are for AbstractJob
                 scheduler=None,
                 label=None, required=None):
        self.messages = messages
        self.sleep = sleep
        self.banner = banner
        super().__init__(label=label, required=required, scheduler=scheduler)

    async def co_run(self):
        """
        Implementation of the method expected by :class:`AbstractJob`
        """
        try:
            if self.banner:
                print(self.banner + " ", end="")
            print(*self.messages)
            if self.sleep:
                print("Sleeping for {}s".format(self.sleep))
                await asyncio.sleep(self.sleep)
        except Exception:                               # pylint: disable=W0703
            # should not happen, but if it does we need to know why
            import traceback
            traceback.print_exc()

    async def co_shutdown(self):
        """
        Implementation of the method expected by :class:`AbstractJob`;
        does nothing.
        """
        pass

    def details(self):                                  # pylint: disable=C0111
        """
        Implementation of the method expected by :class:`AbstractJob`
        """
        result = ""
        if self.sleep:
            result += "[+ sleep {}s] ".format(self.sleep)
        result += "msg= "
        result += self.messages[0]
        result += "..." if len(self.messages) > 1 else ""
        return result



"""
A utility to print time and compute durations, mostly for debugging and tests.
"""

from datetime import datetime

class Watch:
    """
    This class essentially remembers a starting point, so that
    durations relative to that epoch can be printed for debug
    instead of a plain timestamp.

    Parameters:
      message(str): used in the printed message at creation time,
      show_elapsed(bool): tells if a message with the elapsed time needs
        to be printed at creation time (elapsed will be 0),
      show_wall_clock(bool): same for the wall clock.

    Examples:
      Here's a simple use case; note that ``print_wall_clock()``
      is a static because it is mostly useful, precisely, when you
      do not have a ``Watch`` object at hand::

        $ python3
        Python 3.6.4 (default, Mar  9 2018, 23:15:12)
        <snip>
        >>> from asynciojobs import Watch
        >>> import time
        >>> watch = Watch("hello there"); time.sleep(1); watch.print_elapsed()
        000.000  hello there
        001.000  >>>
        >>>
        >>> Watch.print_wall_clock()
        20:48:27.782 >>>
    """

    # default is to print the elapsed format only
    def __init__(self, message=None, *,
                 show_elapsed=True, show_wall_clock=False):
        self.start = 0
        self.reset()
        message = message if message is not None else ""
        if show_elapsed:
            self.print_elapsed(" {}\n".format(message))
        if show_wall_clock:
            self.print_wall_clock(" {}\n".format(message))

    def reset(self):
        """
        Use current wall clock as starting point.
        """
        self.start = datetime.now()

    def seconds(self):
        """
        Returns:
          float: time elapsed since start, in seconds.
        """
        return (datetime.now() - self.start).total_seconds()

    def print_elapsed(self, suffix=" "):
        """
        Print the elapsed time since start in format
        SSS.MMM + a suffix.

        Parameters:
          suffix(str): is appended to the output; to be explicit,
            by default no newline is added.
        """
        print("{:07.3f} {}".format(self.seconds(), suffix),
              end="")

    @staticmethod
    def print_wall_clock(suffix=" "):
        """
        Print current time in HH:MM:SS.MMM + a suffix.

        Parameters:
          suffix(str): is appended to the output; to be explicit,
            by default no newline is added.
        """
        now = datetime.now()
        millisecond = now.microsecond // 1000
        timestamp = datetime.now().strftime("%H:%M:%S")
        print("{}.{}{}".format(timestamp, millisecond, suffix),
              end="")



"""
The ``Scheduler`` class makes it easier to nest scheduler objects.
"""

from asynciojobs import PureScheduler
from asynciojobs import AbstractJob


class Scheduler(PureScheduler, AbstractJob):
    """
    The ``Scheduler`` class is a mixin of the two
    :class:`~asynciojobs.purescheduler.PureScheduler` and
    :class:`~asynciojobs.job.AbstractJob` classes.

    As such it can be used to create nested schedulers,
    since it is a scheduler that can contain jobs,
    and at the same time it is a job, and so it can be included in
    a scheduler.

    Parameters:
      jobs_or_sequences: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`,
        allows to add these jobs inside of the newly-created scheduler;
      jobs_window: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      timeout: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      watch (Watch): passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      verbose (bool): passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      kwds: all other named arguments are sent
        to the :class:`~asynciojobs.job.AbstractJob` constructor.


    Example:

      Here's how to create a very simple scheduler with
      an embedded sub-scheduler; the whole result is equivalent to a simple
      4-steps sequence::

        main = Scheduler(
           Sequence(
             Job(aprint("begin", duration=0.25)),
             Scheduler(
               Sequence(
                 Job(aprint("middle-begin", duration = 0.25)),
                 Job(aprint("middle-end", duration = 0.25)),
               )
             ),
             Job(aprint("end", duration=0.25)),
           )
        main.run()

    Notes:

      There can be several good reasons for using nested schedulers:

      * the scope of a ``window`` object applies to a scheduler, so a nested
        scheduler is a means to apply windoing on a specific set of jobs;
      * likewise the ``timeout`` attribute only applies to the run for the
        whole scheduler;
      * you can use ``forever`` jobs that will be terminated earlier than
        the end of the global scheduler;
      * strictly speaking, the outermost instance in this example could be an
        instance of ``PureScheduler``, but in practice it is simpler to always
        create instances of ``Scheduler``.

      Using an intermediate-level scheduler can in some case help alleviate or
      solve such issues.

    """
    def __init__(self, *jobs_or_sequences,
                 jobs_window=None, timeout=None,
                 watch=None, verbose=False,
                 **kwds):

        PureScheduler.__init__(self, *jobs_or_sequences,
                               jobs_window=jobs_window, timeout=timeout,
                               verbose=verbose, watch=watch)
        AbstractJob.__init__(self, **kwds)

    def _set_sched_id(self, start, id_format):
        """
        Works as a complicit to PureScheduler._set_sched_ids.
        It sets local the ``_sched_id`` attribute and returns the index
        for the next job.
        """
        # first set index on the current (kind of fake) node
        i = AbstractJob._set_sched_id(self,             # pylint: disable=w0212
                                      start, id_format)
        # go on with the jobs in sub scheduler
        return PureScheduler._set_sched_ids(self,       # pylint: disable=w0212
                                            i, id_format)

    def _job_count(self):
        """
        Complicit to PureScheduler._total_length()

        Returns:
          int: 1 + number of nodes included/nested
        """
        return 1 + self._total_length()

    def _list(self, details, depth):
        """
        Complicit to PureScheduler.list()
        """
        print(self.repr_id(),
              self.repr_short(),
              '>'*(depth+1),
              self.repr_main(),
              "{} -> {}".format(self.repr_requires(),
                                self.repr_entries()))
        for job in self.topological_order():
            job._list(details, depth+1)                 # pylint: disable=W0212
        print(self.repr_id(),
              # this should be 7-spaces like repr_short()
              '  end  ',
              '<'*(depth+1),
              self.repr_main(),
              self.repr_exits())

    def dot_cluster_name(self):
        """
        assigns a name to the subgraph that will represent
        a nested scheduler; dot format imposes this name to start
        with ``cluster_`
        """
        return "cluster_{}"\
               .format(self._sched_id)



"""
This module defines the `Sequence` class, that is designed
to ease the building of schedulers
"""


from .job import AbstractJob


class Sequence:
    """
    A Sequence is an object that organizes a set
    of AbstratJobs in a sequence. Its main purpose is to add
    a single `required` relationship per job in the sequence,
    except the for first one, that instead receives as its `required`
    the sequence's requirements.

    If `scheduler` is passed to the sequence's constructor,
    all the jobs passed to the sequence are added in that scheduler.

    Sequences are not first-class citizens, in the sense that
    the scheduler primarily ignores these objects, only the jobs inside
    the sequence matter.

    However a sequence can be used essentially in every place where a job
    could be, either being inserted in an scheduler, added as a
    requirement, and it can have requirements too.

    Parameters:
      sequences_or_jobs: each must be a ``Schedulable`` object,
        the order of course is important here
      required: one, or a collection of, ``Schedulable`` objects that
        will become the requirements for the first job in the sequence
      scheduler: if provided, the jobs in the sequence will be inserted
        in that scheduler.
    """

    def __init__(self, *sequences_or_jobs, required=None, scheduler=None):
        self.jobs = self._flatten(sequences_or_jobs)
        # create the chain of requirements in the sequence
        for job1, job2 in zip(self.jobs, self.jobs[1:]):
            job2.requires(job1)
        # any requirements specified in the constructor
        # actually apply to the first item
        if self.jobs:
            self.jobs[0].requires(required)
        # make all jobs belong in the scheduler if provided
        self.scheduler = scheduler
        if self.scheduler is not None:
            self.scheduler.update(self.jobs)

    @staticmethod
    def _flatten(sequences_or_jobs):
        """
        given an iterable of objects typed either AbstractJob or Sequence
        returns an ordered list of jobs
        """
        result = []
        for joblike in sequences_or_jobs:
            if joblike is None:
                continue
            if isinstance(joblike, AbstractJob):
                result.append(joblike)
            elif isinstance(joblike, Sequence):
                result += joblike.jobs
        return result

    def append(self, *sequences_or_jobs):
        """
        Add these jobs or sequences at the end of the present sequence.

        Parameters:
          sequences_or_jobs: each must be a ``Schedulable`` object.
        """
        if not sequences_or_jobs:
            return
        new_jobs = self._flatten(sequences_or_jobs)
        if self.jobs:
            new_jobs[0].requires(self.jobs[-1])
        self.jobs += new_jobs
        if self.scheduler is not None:
            self.scheduler.update(new_jobs)

    def requires(self, *requirements):
        """
        Adds requirements to the sequence, so that is to say,
        to the first job in the sequence.

        Parameters:
          requirements: each must be a ``Schedulable`` object.

        """
        if not self.jobs:
            # warning ?
            return
        self.jobs[0].requires(*requirements)



"""
Implementation for the Window class, for schedulers
with a limited number of running jobs
"""

import asyncio


class Window:
    """
    The window class implements the logic that allows to
    throttle a scheduler instance to run only a limited number
    of simultaneous jobs.

    Users are not expected to create such objects by themselves,
    the scheduler takes care of that when needed
    """

    def __init__(self, jobs_window, loop):
        # jobs_window needs to be an integer
        if jobs_window is None:
            jobs_window = 0
        self.jobs_window = jobs_window
        self.loop = loop
        self.queue = asyncio.Queue(maxsize=jobs_window, loop=loop)

    def run_job(self, job):
        """
        a decorator around a coroutine, that will first get a slot in the queue

        REMEMBER that this object will need to be CALLED to become
        a future itself
        """
        async def wrapped():                            # pylint: disable=C0111
            # put anything to take a slot in the queue
            await self.queue.put(1)
            job._running = True                         # pylint: disable=w0212
            value = await job.co_run()
            # release slot in the queue
            await self.queue.get()
            # return the right thing
            return value
        return wrapped

    # for debugging
    async def monitor(self, period=3):
        """
        a coroutine that cyclically shows the status of the queue
        until it gets empty
        """
        await asyncio.sleep(period)
        while not self.queue.empty():
            print("queue has {}/{} elements busy"
                  .format(self.queue.qsize(), self.jobs_window))
            await asyncio.sleep(period)



#!/usr/bin/env python3

"""
The PureScheduler class is a set of AbstractJobs, that together with their
*required* relationship, form an execution graph.
"""

import time
import io

import asyncio

from .job import AbstractJob
from .sequence import Sequence
from .window import Window
from .watch import Watch
from .dotstyle import DotStyle

#
# will hopefully go away some day
debug = False                                           # pylint: disable=C0103
# debug = True

# pylint settings
# W0212: we have a lot of accesses to protected members of other classes
# R0914 Too many local variables
# R0904 Too many public methods (xxx this one should be reachable)
# C0302 Too many lines in module (would be nice to not count docstrings)
# pylint: disable=R0914, R0904, C0302

# Historical note: we used to formally define the Schedulable type hint
# but that ended up clobbering the documentation, and was more harmful
# than helpful
# in plain english, a Schedulable object is either
# an instance of AbstractJob or of Sequence


class PureScheduler:                                    # pylint: disable=r0902
    """
    A PureScheduler instance is made of a set of AbstractJob objects.

    The purpose of the scheduler object is to orchestrate an execution of
    these jobs that respects the *required* relationships,
    until they are all complete. It starts with the ones that have no
    requirement, and then triggers the other ones as their requirement
    jobs complete.

    For this reason, the dependency/requirements graph **must be acyclic**.

    Optionnally a scheduler orchestration can be confined to a finite number
    of concurrent jobs (see the ``jobs_window`` parameter below).

    It is also possible to define a ``timeout`` attribute on the object,
    that will limit the execution time of a scheduler.


    Running an AbstractJob means executing its :meth:`co_run()` method,
    which must be a coroutine

    The result of a job's :meth:`~asynciojobs.job.AbstractJob.co_run()`
    is NOT taken into account, as
    long as it returns without raising an exception. If it does raise
    an exception, overall execution is aborted iff the job is
    critical. In all cases, the result and/or exception of each
    individual job can be inspected and retrieved individually at any
    time, including of course once the orchestration is complete.

    Parameters:
      jobs_or_sequences: instances of `AbstractJob` or `Sequence`.
        The order in which they are mentioned is irrelevant.
      jobs_window: is an integer that specifies how many jobs
        can be run simultaneously. None or 0 means no limit.
      timeout: can be an `int` or `float` and is expressed
       in seconds; it applies to the overall orchestration of that scheduler,
       not to any individual job.
      watch: if the caller passes a :class:`~asynciojobs.watch.Watch`
        instance, it is used in debugging messages to show the time
        elapsed wrt that watch, instead of using the wall clock.
      verbose (bool): flag that says if execution should be verbose.


    Examples:
      Creating an empty scheduler::

        s = Scheduler()

      A scheduler with a single job::

        s = Scheduler(Job(asyncio.sleep(1)))

      A scheduler with 2 jobs in parallel::

        s = Scheduler(Job(asyncio.sleep(1)),
                      Job(asyncio.sleep(2)))

      A scheduler with 2 jobs in sequence::

        s = Scheduler(
                Sequence(
                    Job(asyncio.sleep(1)),
                    Job(asyncio.sleep(2))
                ))

    In this document, the ``Schedulable`` name refers to a type hint, that
    encompasses instances of either the `AbstractJob` or `Sequence` classes.

    """

    def __init__(self, *jobs_or_sequences,
                 jobs_window=None, timeout=None,
                 watch=None, verbose=False):

        self.jobs = set(Sequence._flatten(              # pylint: disable=W0212
            jobs_or_sequences))
        self.jobs_window = jobs_window
        # timeout is in seconds
        self.timeout = timeout
        self.watch = watch
        self.verbose = verbose
        # why does it fail ?
        # bool
        self._failed_critical = False
        # False, or the intial timeout
        self._failed_timeout = False
        # see also _record_beginning
        self._expiration = None

    # think of an scheduler as a set of jobs
    def update(self, jobs):
        """
        Adds a collection of ``Schedulable`` objects;
        this method is named after ``set.update()``.

        Parameters:
          jobs: a collection of ``Schedulable`` objects.

        Returns:
          self: the scheduler object, for cascading insertions if needed.
        """
        jobs = set(Sequence._flatten(jobs))             # pylint: disable=W0212
        self.jobs.update(jobs)
        return self

    def add(self, job):
        """
        Adds a single ``Schedulable`` object;
        this method name is inspired from plain python ``set.add()``

        Parameters:
          job: a single ``Schedulable`` object.

        Returns:
          self: the scheduler object, for cascading insertions if needed.
        """
        self.update([job])
        return self

    def __len__(self):
        """
        You can call len() on a PureScheduler object.
        Returns:
          int: number of jobs in the scheduler.
        """
        return len(self.jobs)

    def failed_time_out(self):
        """
        Returns:
          bool: returns True if and only if :meth:`co_run()`
          has failed because of a time out.
        """
        return self._failed_timeout

    def failed_critical(self):
        """
        Returns:
          bool: returns True if and only if :meth:`co_run()`
          has failed because a critical job has raised an exception.
        """
        return self._failed_critical

    def why(self):
        """
        Returns:
          str: a message explaining why :meth:`co_run()` has failed,
          or ``"FINE"`` if it has not failed.

        Notes:

          At this point the code does not check that :meth:`co_run()` has
          actually been called.
        """
        if self._failed_timeout:
            return "TIMED OUT after {}s".format(self._failed_timeout)
        if self._failed_critical:
            return "a CRITICAL job has raised an exception"
        return "FINE"

    ####################
    def sanitize(self):
        """
        This method ensures that the requirements relationship is closed within
        the scheduler. In other words, it removes any requirement attached to a
        job in this scheduler, but that is not itself part of the scheduler.

        This can come in handy in some scheduler whose composition depends on
        external conditions.

        In any case it is crucial that this property holds
        for :meth:`co_run()` to perform properly.

        Returns:
          bool: returns True if scheduler object was fine,
          and False if at least one removal was needed.
        """

        changes = False
        for job in self.jobs:
            before = len(job.required)
            job.required &= self.jobs
            job._s_successors &= self.jobs
            after = len(job.required)
            if before != after:
                changes = True
                if self.verbose:
                    print(20 * '*',
                          "WARNING: job {} has had {} requirements removed"
                          .format(job, before - after))
        return not changes

    ####################
    def rain_check(self):
        """
        Performs a minimal sanity check.
        The purpose of this is primarily to check for cycles,
        and/or missing starting points.

        It's not embedded in :meth:`co_run()` because
        it is not strictly necessary, but it is safer to call this
        before running the scheduler if one wants to double-check the
        jobs dependency graph early on.

        It might also help to have a sanitized scheduler,
        but here again this is up to the caller.

        Returns:
            bool: True if the topology is fine
        """
        try:
            for _ in self.topological_order():
                pass
            return True
        except Exception as exc:                        # pylint: disable=W0703
            if self.verbose:
                print("rain_check failed", exc)
            return False

    ####################
    def topological_order(self):
        """
        A generator function that scans the graph in topological order,
        in the same order as the orchestration,
        i.e. starting from jobs that have no dependencies, and moving forward.

        Beware that this is not a separate iterator, so it can't be nested,
        which in practice should not be a problem.

        Examples:

          Assuming all jobs have a ``label`` attribute,
          print them in the "right" order::

            for job in scheduler.topological_order():
                print(job.label)
        """
        self._reset_marks()
        nb_marked = 0
        target_marked = len(self.jobs)

        while True:
            # detect a fixed point
            changed = False
            # loop on unfinished business
            for job in self.jobs:
                # ignore jobs already marked
                if job._s_mark:                         # pylint: disable=W0212
                    continue
                # if there's no requirement (first pass),
                # or later on if all requirements have already been marked,
                # then we can mark this one
                has_unmarked_requirements = False
                for required_job in job.required:
                    if required_job._s_mark is None:    # pylint: disable=W0212
                        has_unmarked_requirements = True
                if not has_unmarked_requirements:
                    job._s_mark = True                  # pylint: disable=W0212
                    nb_marked += 1
                    changed = True
                    yield job
            # >= is for extra safety but it should be an exact match
            if nb_marked >= target_marked:
                # we're done
                break
            if not changed:
                # this is wrong
                raise Exception(
                    "scheduler could not be scanned"
                    " - most likely because of cycles")
        # if we still have jobs here it's not good either,
        # although it should not happen on a sanitized scheduler
        if nb_marked != target_marked:
            raise Exception("scheduler could not be scanned,"
                            " {} jobs are not reachable from free jobs"
                            .format(target_marked - nb_marked))

    # entry and exit jobs
    def entry_jobs(self):
        """
        A generator that yields all jobs that have no requirement.

        Exemples:

          List all entry points::

             for job in scheduler.entry_points():
                 print(job)
        """
        for job in self.jobs:
            if not job.required:
                yield job

    def exit_jobs(self, *,
                  discard_forever=True,
                  compute_backlinks=True):
        """
        A generator that yields all jobs that are
        not a requirement to another job; it is thus in some sense
        the reverse of :meth:`entry_points()`.

        Parameters:
          discard_forever: if True, jobs marked as forever are skipped; forever
            jobs often have no successors, but are seldom of interest when
            calling this method.
          compute_backlinks: for this method to work properly, it is necessary
            to compute backlinks, an internal structure that holds the opposite
            of the *required* relationship. Passing False here allows to skip
            that stage, when that relationship is known to be up to date
            already.

        """
        if compute_backlinks:
            self._backlinks()
        for job in self.jobs:
            if discard_forever and job.forever:
                continue
            if not job._s_successors:                   # pylint: disable=w0212
                yield job

    def _entry_csv(self):
        result = ", ".join(job.repr_id()
                           for job in self.entry_jobs())
        if result:
            result = "{" + result + "}"
        return result

    def _exit_csv(self, discard_forever=True,
                  compute_backlinks=False):
        result = ", ".join(job.repr_id()
                           for job in self.exit_jobs(
                               discard_forever=discard_forever,
                               compute_backlinks=compute_backlinks))
        if result:
            result = "{" + result + "}"
        return result

    def some_entry_job(self):
        """
        Returns one randomly picked entry job
        Needed when creating the dot format, because
        of the specifics of that format; although it may be
        that returning any job would do the job as well
        """
        for job in self.entry_jobs():
            return job
        raise ValueError("Internal error")

    def some_exit_job(self):
        """
        Same for exit nodes;
        """
        for job in self.exit_jobs():
            return job
        raise ValueError("Internal error")

    def repr_entries(self):                             # pylint: disable=c0111
        return "entries={}".format(self._entry_csv())

    def repr_exits(self):                               # pylint: disable=c0111
        return "exits={}".format(self._exit_csv(compute_backlinks=True))

    ####################
    def _reset_marks(self):
        """
        reset Job._s_mark on all jobs
        """
        for job in self.jobs:
            job._s_mark = None                          # pylint: disable=W0212

    def _reset_tasks(self):
        """
        In case one tries to run the same scheduler twice
        """
        for job in self.jobs:
            job._task = None                            # pylint: disable=W0212

    def _backlinks(self):
        """
        initialize Job._s_successors on all jobs
        as the reverse of Job.required
        """
        for job in self.jobs:
            job._s_successors = set()                   # pylint: disable=W0212
        for job in self.jobs:
            for req in job.required:
                req._s_successors.add(job)              # pylint: disable=W0212

    def _ensure_future(self, job, window, loop):        # pylint: disable=R0201
        """
        this is the hook that lets us make sure the created Task objects
        have a backlink pointer to their corresponding job
        """
        #
        # this is where we call co_run()
        #
        # the decorated object is a coroutine that needs to be CALLED:
        #                                               vv
        task = asyncio.ensure_future(window.run_job(job)(), loop=loop)
        # create references back and forth between Job and asyncio.Task
        task._job = job                                 # pylint: disable=W0212
        job._task = task                                # pylint: disable=W0212
        return task

    def _record_beginning(self, timeout):
        """
        Called once at the beginning of :meth:`co_run()`, this method
        computes the absolute expiration date when a timeout is defined.
        """
        self._expiration = \
            None if timeout is None \
            else time.time() + timeout

    def _remaining_timeout(self):
        """
        Called each time :meth:`co_run()` is about to call `asyncio.wait()`,
        this method computes the timeout argument for wait
        - or None if co_run is called without a timeout
        """
        return \
            None if self._expiration is None \
            else self._expiration - time.time()

    async def _tidy_tasks(self, pending):
        """
        Once :meth:`co_run()` is done, in order to tidy up the underlying
        Task objects that have not completed, it is necessary to cancel
        them and wait for them.

        According to the context, this can be with forever tasks,
        or because a timeout has occured.
        """
        if pending:
            for task in pending:
                task.cancel()
            # wait for the forever tasks for a clean exit
            # don't bother to set a timeout, as this is expected
            # to be immediate since all tasks are canceled
            await asyncio.wait(pending)

    async def _tidy_tasks_exception(self, tasks):
        """
        Similar but in order to clear the exceptions,
        we need to run gather() instead
        """
        # do not use task._job.raised_exception()
        # so we can use this with co_shutdown() tasks as well
        # (these are not attached to a job)
        exception_tasks = [task for task in tasks
                           if task._exception]          # pylint: disable=W0212
        for task in exception_tasks:
            task.cancel()
            # if debug is turned on, provide details on the exceptions
            if debug:
                job = task._job                         # pylint: disable=W0212
                self._show_task_stack(
                    task,
                    "TIDYING {} {} {}"
                    .format(job.repr_id(), job.repr_short(), job.repr_main()))
        # don't bother to set a timeout,
        # this is expected to be immediate
        # since all tasks are canceled
        await asyncio.gather(*exception_tasks, return_exceptions=True)

    @staticmethod
    def _show_task_stack(task, msg='STACK', margin=4, limit=None):
        if isinstance(task, AbstractJob):
            task = task._task                           # pylint: disable=W0212
        sep = margin * ' ' + 20 * '*'
        print(sep)
        print(sep, 'BEG ' + msg)
        print(sep)
        # naive approach would be like this, but does not support margin:
        # task.print_stack()
        stio = io.StringIO()
        task.print_stack(file=stio, limit=limit)
        stio.seek(0)
        for line in stio:
            print(margin * ' ' + line, end="")
        print(sep)
        print(sep, 'END ' + msg)
        print(sep)

    async def co_shutdown(self):
        """
        Shut down the scheduler, by sending a message to all the jobs when
        orchestration is over.

        Typically for example, several jobs sharing the same ssh connection
        will arrange for that connection to be kept alive across an entire
        scheduler lifespan, but there is a need to tear these connections
        down eventually.

        Returns:
          None
        """

        await self._feedback(None, "scheduler is shutting down...")
        tasks = [asyncio.ensure_future(job.co_shutdown())
                 for job in self.jobs]
        # the done part is of no use here
        _, pending = await asyncio.wait(
            tasks, timeout=self._remaining_timeout())
        if pending:
            print("WARNING: {}/{} co_shutdown() methods"
                  " have not returned within timeout"
                  .format(len(pending), len(self.jobs)))
            await self._tidy_tasks(pending)
        # xxx should consume any exception as well ?
        # self._tidy_tasks_exception(done)

    async def _feedback(self, jobs, state, force=False):
        """
        When self.verbose is set, provide feedback about the mentioned
        jobs having reached this state
        if jobs is None, then state is a message to be shown as-is
        jobs may be a collection or an individual Job or Task object
        """
        if not force and not self.verbose:
            return

        def print_time():                               # pylint: disable=c0111
            if self.watch is not None:
                self.watch.print_elapsed()
            else:
                Watch.print_wall_clock()
        name = ""
        # with this line we allow a PureScheduler to have a label
        # it's mostly applicable to Scheduler instances, but can come in
        # handy for pure Scheduler's as well
        if hasattr(self, 'label'):
            name = "({})".format(self.label)            # pylint: disable=e1101

        # general feedback when no job is specified by caller
        if jobs is None:
            print_time()
            print("SCHEDULER{}: {}".format(name, state))
            return
        if not isinstance(jobs, (list, set, tuple)):
            jobs = (jobs,)
        for job in jobs:
            if not isinstance(job, AbstractJob):
                # we expect a task here
                job = job._job                          # pylint: disable=W0212
            print_time()
            print("{} {}: {} {} {}"
                  .format(name, state,
                          job.repr_id(), job.repr_short(), job.repr_main()),
                  end="")
            if self.verbose:
                print(" {} {}"
                      .format(job.repr_result(),
                              job.repr_requires()),
                      end="")
            print()

    ####################
    def run(self, *args, loop=None, **kwds):
        """
        A synchroneous wrapper around :meth:`co_run()`,
        please refer to that link for details on parameters and return value.

        Also, the canonical name for this is ``run()`` but for historical
        reasons you can also use ``orchestrate()`` as an alias for ``run()``.
        """
        if loop is None:
            loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.co_run(loop=loop, *args, **kwds))

    # define the alias for legacy
    orchestrate = run

    async def co_run(self, loop=None):            # pylint: disable=R0912,R0915

        """
        The primary entry point for running a scheduler.
        See also :meth:`run()` for a synchronous wrapper around this coroutine.

        Runs member jobs (that is, schedule their `co_run()` method)
        in an order that satisfies their *required* relationsship.

        Proceeds to the end no matter what, except if either:

        1. one critical job raises an exception, or

        2. a timeout occurs.

        Returns:
          bool: `True` if none of these 2 conditions occur, `False` otherwise.

        Jobs marked as ``forever`` are not waited for. All jobs get
        terminated through their `co_shutdown()` method.

        Parameters:
          loop: an asyncio event loop, it defaults to
            ``asyncio.get_event_loop()``, which is almost certainly
            the right value to use.

        """
        if loop is None:
            loop = asyncio.get_event_loop()
        # create a Window no matter what; it will know what to do
        # also if jobs_window is None
        window = Window(self.jobs_window, loop)

        # initialize; this one is not crucial but is helpful
        # for debugging purposes
        self._set_sched_ids()
        # backlinks - i.e. _s_successors is the reverse of required
        self._backlinks()
        # clear any Task instance
        self._reset_tasks()
        # for computing global timeout
        self._record_beginning(self.timeout)
        # reset status
        self._failed_critical = False
        self._failed_timeout = False

        # empty schedulers are fine too
        if not self.jobs:
            return True

        # how many jobs do we expect to complete: the ones that don't run
        # forever
        nb_jobs_finite = len([j for j in self.jobs if not j.forever])
        # the other ones
        nb_jobs_forever = len(self.jobs) - nb_jobs_finite
        # count how many jobs have completed
        nb_jobs_done = 0

        # start with the free jobs
        entry_jobs = [job for job in self.jobs if not job.required]

        if not entry_jobs:
            raise ValueError("No entry points found - cannot orchestrate")

        if self.verbose:
            await self._feedback(None, "entering co_run() with {} jobs"
                                 .format(len(self.jobs)))

        await self._feedback(entry_jobs, "STARTING")

        pending = [self._ensure_future(job, window, loop=loop)
                   for job in entry_jobs]

        while True:
            done, pending \
                = await asyncio.wait(pending,
                                     timeout=self._remaining_timeout(),
                                     return_when=asyncio.FIRST_COMPLETED)

            done_ok = {t for t in done
                       if not t._exception}             # pylint: disable=W0212
            await self._feedback(done_ok, "DONE")
            done_ko = done - done_ok
            await self._feedback(done_ko, "RAISED EXC.")

            # nominally we have exactly one item in done
            # the only condition where we have nothing in done is
            # because a timeout occurred
            # there are also cases where done has more than one entry
            # typically when 2 jobs have very similar durations
            if not done:
                await self._feedback(None,
                                     "PureScheduler.co_run: TIMEOUT occurred",
                                     force=True)
                # clean up
                await self._feedback(pending, "ABORTING")
                await self._tidy_tasks(pending)
                await self.co_shutdown()
                self._failed_timeout = self.timeout
                return False

            # exceptions need to be cleaned up
            # clear the exception(s) in done
            await self._tidy_tasks_exception(done)
            # do we have at least one critical job with an exception ?
            critical_failure = False
            for done_task in done:
                done_job = done_task._job               # pylint: disable=W0212
                if done_job.raised_exception():
                    critical_failure = critical_failure \
                        or done_job.is_critical()
                    await self._feedback(
                        done_job, "EXCEPTION occurred - on {}critical job"
                        .format("non-" if not done_job.is_critical() else ""))
                    # make sure these ones show up even if not in debug mode
                    if debug:
                        self._show_task_stack(done_task, "DEBUG")
            if critical_failure:
                await self._tidy_tasks(pending)
                await self.co_shutdown()
                self._failed_critical = True
                await self._feedback(
                    None, "Emergency exit upon exception in critical job",
                    force=True)
                return False

            # are we done ?
            # only account for not forever jobs (that may still finish, one
            # never knows)
            done_jobs_not_forever = {
                j for j in done
                if not j._job.forever}                  # pylint: disable=W0212
            nb_jobs_done += len(done_jobs_not_forever)

            if nb_jobs_done == nb_jobs_finite:
                if debug:
                    print("PureScheduler.co_run: {} CLEANING UP at iter. {}/{}"
                          .format(4 * '-', nb_jobs_done, nb_jobs_finite))
                if self.verbose and nb_jobs_forever != len(pending):
                    print("WARNING - apparent mismatch"
                          " - {} forever jobs, {} are pending"
                          .format(nb_jobs_forever, len(pending)))
                await self._feedback(pending, "TIDYING forever")
                await self._tidy_tasks(pending)
                await self.co_shutdown()
                return True

            # go on : find out the jobs that can be added to the mix
            # only consider the ones that are right behind any of the the jobs
            # that just finished
            possible_next_jobs = set()
            for done_task in done:
                possible_next_jobs.update(
                    done_task._job._s_successors)       # pylint: disable=W0212

            # find out which ones really can be added
            added = 0
            for candidate_next in possible_next_jobs:
                # do not add an job twice
                if candidate_next.is_running():
                    continue
                # we can start only if all requirements are satisfied
                # at this point entry points have is_running() -> return True
                # so they won't run this code
                requirements_ok = True
                for req in candidate_next.required:
                    if not req.is_done():
                        requirements_ok = False
                if requirements_ok:
                    await self._feedback(candidate_next, "STARTING")
                    pending.add(self._ensure_future(
                        candidate_next, window, loop=loop))
                    added += 1

    def _total_length(self):
        """
        Counts the total number of jobs that need to be numbered
        in nested scenarii.

        A regular job counts for 1,
        and a scheduler counts for 1 + its own _total_length

        Returns:
          int: total number of nodes in subject and nested schedulers

        """
        return sum(job._job_count()                     # pylint: disable=w0212
                   for job in self.jobs)

    def _set_sched_ids(self, start=1, id_format=None):
        """
        Write into each job._sched_id an id compliant
        with topological order.

        Returns:
          int: the next index to use
        """
        # id_format is computed once by the toplevel scheduler
        # and then passed along the tree

        if id_format is None:
            import math
            # how many chars do we need to represent all jobs
            total = self._total_length()
            width = 1 if total <= 9 \
                else int(math.log(total-1, 10)) + 1
            # id_format is intended to be e.g. {:02d}
            id_format = "{{:0{w}d}}".format(w=width)    # pylint: disable=w1303
        i = start
        for job in self.topological_order():
            i = job._set_sched_id(i, id_format)         # pylint: disable=w0212
        return i

    ####################
    def list(self, details=False):
        """
        Prints a complete list of jobs in topological order, with their status
        summarized with a few signs. See the README for examples and a legend.

        Beware that this might raise an exception if :meth:`rain_check()`
        would return ``False``, i.e. if the graph is not acyclic.
        """
        # so now we can refer to other jobs by their id when showing
        # requirements
        self._set_sched_ids()
        for job in self.topological_order():
            job._list(details, 0)                       # pylint: disable=W0212

    def list_safe(self):
        """
        Print jobs in no specific order, the advantage being that it
        works even if scheduler is broken wrt :meth:`rain_check()`
        """
        for i, job in enumerate(self.jobs):
            print(i, job)

    def __repr__(self):
        # linter says unused variable but it is indeed used in f-string
        nb_total = len(self.jobs)                       # pylint: disable=W0612
        done = {j for j in self.jobs if j.is_done()}
        nb_done = len(done)                             # pylint: disable=W0612
        running = {j for j in self.jobs if j.is_running()}
        ongoing = running - done
        nb_ongoing = len(ongoing)                       # pylint: disable=W0612
        idle = self.jobs - running
        nb_idle = len(idle)                             # pylint: disable=W0612
        return "{type} with {nb_done} done + {nb_ongoing} ongoing" \
               " + {nb_idle} idle = {nb_total} job(s)" \
            .format(type=type(self).__name__, **locals())

    def debrief(self, details=False):
        """
        Designed for schedulers that have failed to orchestrate.

        Print a complete report, that includes `list()` but also gives
        more stats and data.
        """
        nb_total = len(self.jobs)
        done = {j for j in self.jobs if j.is_done()}
        nb_done = len(done)
        running = {j for j in self.jobs if j.is_running()}
        ongoing = running - done
        nb_ongoing = len(ongoing)
        idle = self.jobs - running
        nb_idle = len(idle)

        exceptions = {j for j in self.jobs if j.raised_exception()}
        criticals = {j for j in exceptions if j.is_critical()}

        message = "scheduler has a total of {} jobs".format(nb_total)

        def legible_message(number, adj):               # pylint: disable=C0111
            if number == 0:
                return " none is {}".format(adj)
            if number == 1:
                return " 1 is {}".format(adj)
            return " {} are {}".format(number, adj)
        message += ", " + legible_message(nb_done, "done")
        message += ", " + legible_message(nb_ongoing, "ongoing")
        message += ", " + \
            legible_message(nb_idle, "idle (or scheduled but not running)")

        print(5 * '-', self.why())
        self.list(details)
        #####
        if exceptions:
            nb_exceptions = len(exceptions)
            nb_criticals = len(criticals)
            print("===== {} job(s) with an exception, including {} critical"
                  .format(nb_exceptions, nb_criticals))
            # show critical exceptions first
            for j in self.topological_order():
                if j in criticals:
                    self._show_task_stack(
                        j, "stack for CRITICAL JOB {} {} {}"
                        .format(j.repr_id(), j.repr_short(), j.repr_main()))
            # then exceptions that were not critical
            non_critical_exceptions = exceptions - criticals
            for j in self.topological_order():
                if j not in non_critical_exceptions:
                    continue
                if not self.verbose:
                    print("non-critical: {}: exception {}"
                          .format(j._get_text_label(),  # pylint: disable=W0212
                                  j.raised_exception()))
                    if self.verbose:
                        self._show_task_stack(
                            j, "non-critical job exception stack")

    # graphical outputs
    #
    # in a first attempt we had one function to store a dot format into a file
    # and another one to build the graph natively; hence duplication of code
    # there are simpler means to do that
    # in addition with nested schedulers things become a bit messy, so
    # it's crucial to stick to one single code

    def dot_format(self):
        """
        Creates a graph that depicts the jobs and their *requires*
        relationships, in `DOT Format`_.

        Returns:
          str: a representation of the graph in `DOT Format`_
          underlying this scheduler.

        See graphviz_'s documentation, together with its `Python wrapper
        library`_, for more information on the format and available tools.

        See also `Wikipedia on DOT`_ for a list of tools that support the
        ``dot`` format.



        As a general rule, ``asynciojobs`` has a support for producing
        `DOT Format`_ but stops short of actually importing ``graphviz``
        that can be cumbersome to install, but for the notable exception
        of the :meth:graph() method. See that method for how
        to convert a ``PureScheduler`` instance into a native ``DiGraph``
        instance.

        .. _DOT Format: https://graphviz.gitlab.io/_pages/doc/info/lang.html

        .. _graphviz: https://graphviz.gitlab.io/documentation/

        .. _Python wrapper library: https://graphviz.readthedocs.io/en/stable/

        .. _Wikipedia on DOT: https://en.wikipedia.org/wiki/
DOT_%28graph_description_language%29
        """
        self._set_sched_ids()
        # xxx should maybe use/show scheduler's label if set
        return "digraph asynciojobs" + self._dot_body(DotStyle())

    def _dot_body(self, dot_style):
        """
        Creates the dot body for a scheduler, i.e the part between
        brackets, without the surrounding ``digraph`` or ``subgraph``
        declaration, that must be added from the outside, depending
        on whether we have a main scheduler or a nested one.
        """
        # use ids so as to not depend on labels

        result = ""
        result += "{\n"
        result += "compound=true;\n"
        result += "graph [{}];\n".format(dot_style)
        for job in self.topological_order():

            # regular jobs
            if not isinstance(job, PureScheduler):
                # declare node, attach label, and set visual attributes
                result += job.repr_id()
                result += ' [{}]\n'.format(job.dot_style())

                # add edges
                for req in job.required:

                    # upstream is a regular job
                    if not isinstance(req, PureScheduler):
                        result += ("{} -> {};\n"
                                   .format(req.repr_id(), job.repr_id()))

                    # upstream is a scheduler
                    else:
                        from_node = req.some_exit_job()
                        cluster_name = req.dot_cluster_name()
                        result += ("{} -> {} [ltail={}];\n"
                                   .format(from_node.repr_id(),
                                           job.repr_id(),
                                           cluster_name))

            # nested scheduler
            else:
                # xxx should maybe use/show scheduler's label if set
                # insert a subgraph instead

                cluster_name = job.dot_cluster_name()
                result += "subgraph {}".format(cluster_name)
                result += job._dot_body(                # pylint: disable=w0212
                    job.dot_style())

                for req in job.required:

                    # upstream is a regular job
                    if not isinstance(req, PureScheduler):
                        result += ("{} -> {} [lhead={}];\n"
                                   .format(req.repr_id(),
                                           job.some_entry_job().repr_id(),
                                           cluster_name))

                    # upstream is a scheduler as well
                    else:
                        src_cluster_name = req.dot_cluster_name()
                        result += ("{} -> {} [lhead={} ltail={}];\n"
                                   .format(req.some_exit_job().repr_id(),
                                           job.some_entry_job().repr_id(),
                                           cluster_name,
                                           src_cluster_name))

        result += "}\n"
        return result

    def export_as_dotfile(self, filename):
        """
        This method does not require ``graphviz`` to be installed, it
        writes a file in dot format for post-processing with
        e.g. graphviz's ``dot`` utility. It is a simple wrapper around
        :meth:`dot_format()`.

        Parameters:
          filename: where to store the result.

        Returns:
          str: a message that can be printed for information, like e.g.
          ``"(Over)wrote foo.dot"``

        See also the :meth:`graph()` method that serves a similar purpose but
        natively as a ``graphviz`` object.

        As an example of post-processing, a PNG image can be then obtained from
        that dotfile with e.g.::

          dot -Tpng foo.dot -o foo.png
        """
        with open(filename, 'w') as output:
            output.write(self.dot_format())
        return "(Over)wrote {}".format(filename)

    def graph(self):
        """
        Returns:
          graphviz.Digraph: a graph

        This method serves the same purpose as :meth:`export_to_dotfile()`,
        but it natively returns a graph instance. For that reason,
        its usage requires the installation of the ``graphviz`` package.

        This method is typically useful in a Jupyter notebook,
        so as to visualize a scheduler in graph format  - see
        http://graphviz.readthedocs.io/en/stable/manual.html#jupyter-notebooks
        for how this works.

        The dependency from ``asynciojobs`` to ``graphviz`` is limited
        to this method, as it is the only place that needs it,
        and as installing ``graphviz`` can be cumbersome.

        For example, on MacOS I had to do both::

          brew install graphviz     # for the C/C++ binary stuff
          pip3 install graphviz     # for the python bindings
        """

        from graphviz import Source
        return Source(source=self.dot_format())



#!/usr/bin/env python3

"""
package / install
"""

from sys import version_info

import setuptools

from asynciojobs import __version__

# check python version
MAJOR, MINOR = version_info[0:2]
if not (MAJOR == 3 and MINOR >= 5):
    print("python 3.5 or higher is required")
    exit(1)


LONG_DESCRIPTION = \
                   "See notebook at https://github.com/parmentelat/" \
                   "asynciojobs/blob/master/README.ipynb"

REQUIRED_MODULES = []

setuptools.setup(
    name="asynciojobs",
    version=__version__,
    author="Thierry Parmentelat",
    author_email="thierry.parmentelat@inria.fr",
    description="A simplistic orchestration engine for asyncio-based jobs",
    long_description=LONG_DESCRIPTION,
    license="CC BY-SA 4.0",
    url="http://asynciojobs.readthedocs.io/",
    packages=['asynciojobs'],
    install_requires=REQUIRED_MODULES,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3.5",
    ],
)



# -*- coding: utf-8 -*-

"""
This module defines :class:`AbstractJob`, that is the base class
for all the jobs in a Scheduler, as well as a basic concrete subclass
:class:`Job` for creating a job from a coroutine.

It also defines a couple of simple job classes.
"""

import sys
import asyncio

from .dotstyle import DotStyle

debug = False                                           # pylint: disable=C0103
debug = True                                            # pylint: disable=C0103

# pylint settings
# W0212: we have a lot of accesses to protected members of other classes
# R0914 Too many local variables
# pylint: disable=W0212


# my first inclination had been to specialize asyncio.Task
# it does not work well though, because you want to model
# dependencies **before** anything starts, of course
# but in asyncio, creating a Task object implies scheduling that for execution

# so, let's have it the other way around
# what we need is a way to attach our own Job instance to the corresp.
# Task instance (and back) right after Task creation, so that
# (*) once asyncio.wait is done, we can easily find out
#     wich Jobs are done or pending
# (*) from one Job, easily know what its status is by
#     looking into its Task obj - if already scheduled

# Scheduler == graph
# Job == node


class AbstractJob:                                      # pylint: disable=R0902
    """
    AbstractJob is a virtual class:

    * it offers some very basic graph-related features
      to model requirements *a la* Makefile;
    * its subclasses are expected to implement a `co_run()`
      and a `co_shutdown()` methods that specify
      the actual behaviour of the job, as coroutines.
    * AbstractJob is mostly a companion class to the
      :class:`~asynciojobs.purescheduler.PureScheduler` class,
      that triggers these co_* methods.

    **Life Cycle**: AbstractJob is also aware of a common life cycle
    for all jobs, which can be summarized as follows:

      **idle** → **scheduled** → **running** → **done**

    In un-windowed schedulers, there is no distinction between
    scheduled and running. In other words, in this case a job goes directly
    from **idle** to **running**.a

    On the other hand, in windowed orchestrations - see the ``jobs_window``
    attribute to
    :meth:`~asynciojobs.purescheduler.PureScheduler` -
    a job can be scheduled but not yet running, because it is waiting
    for a slot in the global window.

    Args:

        forever (bool): if set, means the job
          is not returning at all and runs forever;
          in this case ``Scheduler.orchestrate()``
          will not wait for that job, and will terminate it
          once all the regular - i.e. not-forever - jobs are done.

        critical (bool): if set,
          this flag indicates that any exception raised during the
          execution of that job should result in the scheduler
          aborting its run immediately. The default behaviour
          is to let the scheduler finish its jobs, at which point
          the jobs can be inspected for exceptions or results.

        required: this can be one, or a collection of, jobs that will
          make the job's requirements;
          requirements can be added later on as well.

        label (str): for convenience mostly, allows to specify
          the way that particular job should be displayed by the scheduler,
          either in textual form by ``Scheduler.list()``, or in graphical form
          by ``Scheduler.graph()``. See also :meth:`text_label()`
          and :meth:`graph_label()` for how this is used.

          As far as labelling, each subclass of :class:`AbstractJob`
          implements a default labelling scheme, so it is not mandatory
          to set a specific label on each job instance, however it is
          sometimes useful.

          Labels must not be confused with details, see :meth:`details()`

        scheduler: this can be an instance of a
          :class:`~asynciojobs.purescheduler.PureScheduler` object, in which
          the newly created job instance is immediately added.
          A job instance can also be inserted in a scheduler instance later on.

    **Note**: a Job instance must only be added in **one Scheduler instance**
    at most - be aware that the code makes no control on this property,
    but be aware that odd behaviours can be observed if it is not fulfilled.
    """

    def __init__(self, *,                               # pylint: disable=R0913
                 forever=False, critical=False, label=None,
                 required=None, scheduler=None):
        self.forever = forever
        self.critical = critical
        # access to labelling is done through methods
        self.label = label
        # for convenience, one can mention
        # only one, or a collection of, AbstractJobs
        self.required = set()
        self.requires(required)
        # convenience again
        if scheduler is not None:
            scheduler.add(self)
        # once submitted in the asyncio loop/scheduler,
        # `co_run()` gets embedded in a Task object,
        # that is our handle when talking to asyncio.wait
        self._task = None
        # this is updated by the Window class when the job makes it through
        self._running = False
        # ==== fields for our friend Scheduler all start with _s_
        # this is for graph browsing algos
        self._s_mark = None
        # the reverse of required
        self._s_successors = set()
        # this attribute is reserved for use by the scheduler
        # that will for example store there an ordering information,
        # which in turn can be used for printing relationships
        # by Scheduler.list() and similar
        self._sched_id = None

    def _set_sched_id(self, start, id_format):
        """
        Works as a complicit to Scheduler._set_sched_ids.
        It sets local the ``_sched_id`` attribute and returns the index
        for the next job.

        This is defined as a method so that (nestable) Scheduler
        can redefine it.

        Returns:
          int: next index, in this case ``start + 1``
        """
        self._sched_id = str(id_format.format(start))
        return start + 1

    def _job_count(self):                               # pylint: disable=r0201
        """
        Complicit to Scheduler._total_length()
        """
        return 1

    # we don't use depth
    def _list(self, details, _):
        """
        Complicit to Scheduler.list()
        """
        print(self.repr_id(),
              self.repr_short(),
              self.repr_main(),
              self.repr_result(),
              self.repr_requires())
        if details and hasattr(self, 'details'):
            details = self.details()
            if details is not None:
                print(details)

    def _get_text_label(self):
        # In terms of labelling, things have become a little tricky over
        # time. When listing an instance of Scheduler, there are 2 ways
        # we need to show a job
        #
        # * first there is a plain label, that may be set at creation time
        #
        # * second, when showing references (like the jobs that a given job
        #  requires), we show ids like '01' and similar.
        # Except that, the job itself has no idea about that at first,
        # it's the Scheduler instance that decides on that.
        #

        # use instance-specific label if set
        attempt = self.label
        if attempt is not None:
            return attempt
        # otherwise, try self.text_label() and use that
        attempt = self.text_label()
        if attempt is not None:
            return attempt
        # otherwise
        return "NOLABEL"

    def _get_graph_label(self):
        # a similar logic for a graphical label
        # we don't need to bother about _sched_id here though
        # also we try graph_label() first, and resort to text_label()
        # if it's not redefined on the object
        attempt = self.graph_label()
        if attempt is not None:
            return attempt
        return self._get_text_label()

    def text_label(self):
        """
        This method is intended to be redefined by daughter classes.

        Returns:
          a one-line string that describes this job.

        This representation for the job is used by the Scheduler object
        through its :meth:`~asynciojobs.purescheduler.PureScheduler.list()` and
        :meth:`~asynciojobs.purescheduler.PureScheduler.debrief()` methods,
        i.e. when a scheduler is printed out in textual format.

        The overall logic is to always use the instance's ``label`` attribute
        if set, or to use this method otherwise. If none of this returns
        anything useful, the textual label used is ``NOLABEL``.
        """
        pass

    def graph_label(self):
        """
        This method is intended to be redefined by daughter classes.

        Returns:
          a string used by the Scheduler methods
          that produce a graph, such as
          :meth:`~asynciojobs.purescheduler.PureScheduler.graph` and
          :meth:`~asynciojobs.purescheduler.PureScheduler.export_as_dotfile`.

        Because of the way graphs are presented, it can have contain "newline"
        characters, that will render as line breaks in the output graph.

        If this method is not defined on a concrete class,
        then the :meth:`text_label()` method is used instead.
        """
        pass

    def dot_style(self):                                # pylint: disable=c0111
        """
        This method computes the DOT attributes
        which are used to style boxes according to
        critical / forever / and similar.

        Legend is quite simply that:

        * critical jobs have a red border, and
        * the border of forever jobs is made of a dashed line.

        Returns:
          DotStyle: a dict-like mapping that sets DOT
          attributes for that job.
        """
        # see also https://graphviz.gitlab.io/_pages/doc/info/attrs.html
        data = DotStyle()
        # style; DotStyle known how to deal with lists
        data['style'] = []
        # label
        label = self._get_graph_label()
        if label != 'NOLABEL':
            data['label'] = label
        # common attributes : rounded box
        data['shape'] = 'box'
        data['style'].append('rounded')
        # forever items have a dashed line
        if self.forever:
            data['style'].append('dashed')
        # critical elements have a red border
        if self.is_critical():
            data['color'] = 'red'
        return data

    ##########
    _has_support_for_unicode = None  # type: bool

    @classmethod
    def _detect_support_for_unicode(cls):
        if cls._has_support_for_unicode is None:
            try:
                cls._c_saltire.encode(sys.stdout.encoding)
                cls._has_support_for_unicode = True
            except UnicodeEncodeError:
                cls._has_support_for_unicode = False
        return cls._has_support_for_unicode

    # unicode version
    # _c_frowning_face = "\u2639" # ☹
    # _c_smiling_face  = "\u263b" # ☻
    _c_saltire = "\u2613"       # ☓
    _c_circle_arrow = "\u21ba"  # ↺
    _c_black_flag = "\u2691"    # ⚑
    _c_white_flag = "\u2690"    # ⚐
    _c_warning = "\u26a0"       # ⚠
    _c_black_star = "\u2605"    # ★
    _c_sun = "\u2609"           # ☉
    _c_infinity = "\u221e"      # ∞

    def _short_unicode(self):
        """
        A small (7 chars) badge that summarizes the job's internal attributes
        uses non-ASCII characters
        """
        # where is it in the lifecycle
        c_running = self._c_saltire if self.is_done() else \
            self._c_circle_arrow if self.is_running() else \
            self._c_black_flag if self.is_scheduled() else \
            self._c_white_flag
        # is it critical or not ?
        c_crit = self._c_warning if self.is_critical() else " "
        # has it raised an exception or not ?
        c_boom = self._c_black_star if self.raised_exception() \
            else self._c_sun if self.is_running() \
            else " "
        # is it going forever or not
        c_forever = self._c_infinity if self.forever else " "

        # add extra white space as unicode chars in terminal tend to be wider
        # than others
        return "{} {} {} {}".format(c_crit, c_boom, c_running, c_forever)

    def _short_ascii(self):
        """
        A small (7 chars) badge that summarizes the job's internal attributes
        uses ASCII-only characters
        """
        # where is it in the lifecycle
        c_running = "x" if self.is_done() else \
            "o" if self.is_running() else \
            "." if self.is_scheduled() else \
            ">"
        # is it critical or not ?
        c_crit = "!" if self.is_critical() else " "
        # has it raised an exception or not ?
        c_boom = ":(" if self.raised_exception() \
                 else ":)" if self.is_running() \
                 else "  "
        # is it going forever or not
        c_forever = "8" if self.forever else " "

        # add extra white space as unicode chars in terminal tend to be wider
        # than others
        return "{} {} {} {}".format(c_crit, c_boom, c_running, c_forever)

    def repr_id(self):
        """
        Returns:
          str: the job's id inside the scheduler, or '??' if that was
          not yet set by the scheduler.
        """
        return self._sched_id or '??'

    def repr_short(self):
        """
        Returns:
          str: a 4 characters string (in fact 7 with interspaces)
            that summarizes the 4 dimensions of the job, that is to say

        * its point in the lifecycle (idle → scheduled → running → done)

        * is it declared as forever

        * is it declared as critical

        * did it trigger an exception

        """
        if self._detect_support_for_unicode():
            return self._short_unicode()
        return self._short_ascii()

    def _req_csv(self):
        if not self.required:
            return ""
        return ("{"
                + ", ".join(req.repr_id()
                            for req in self.required)
                + "}")

    def repr_main(self):
        """
        Returns:
          str: standardized body of the object's repr,
            like e.g. ``<SshJob `my command`>``.
        """
        return ("<{} `{}`>"
                .format(type(self).__name__,
                        self._get_text_label()))

    def repr_result(self):
        """
        Returns:
          str: standardized repr's part that shows
            the result or exception of the job.
        """
        exception = self.raised_exception()
        if exception:
            critical_msg = "CRIT. EXC." if self.is_critical() \
                           else "exception"
            return ("!! {} => {}:{}!!"
                    .format(critical_msg,
                            type(exception).__name__, exception))
        if self.is_done():
            return "[[ -> {}]]".format(self.result())
        return "[not done]"

    def repr_requires(self):
        """
        Returns:
          str: text part that describes requirements
        """
        if self.required:
            return "requires={}".format(self._req_csv())
        return ""

    def __repr__(self):
        return (
            "{} {} {} {}"
            .format(self.repr_short(),
                    self.repr_main(),
                    self.repr_result(),
                    self.repr_requires()))

    def requires(self, *requirements):
        """
        Arguments:
          requirements: an iterable of `AbstractJob`
            instances that are added to the requirements.

        For convenience, any nested structure made of job instances
        can be provided, and if None objects are found, they are silently
        ignored. For example, with `j{1,2,3,4}` being jobs or sequences,
        all the following calls are legitimate:

        * ``j1.requires(None)``
        * ``j1.requires([None])``
        * ``j1.requires((None,))``
        * ``j1.requires(j2)``
        * ``j1.requires(j2, j3)``
        * ``j1.requires([j2, j3])``
        * ``j1.requires(j2, [j3, j4])``
        * ``j1.requires((j2, j3))``
        * ``j1.requires(([j2], [[[j3]]]))``
        """
        from .sequence import Sequence
        for requirement in requirements:
            if requirement is None:
                continue
            if isinstance(requirement, AbstractJob):
                self.required.add(requirement)
            elif isinstance(requirement, Sequence):
                if requirement.jobs:
                    self.required.add(requirement.jobs[-1])
            elif isinstance(requirement, (tuple, list)):
                for req in requirement:
                    self.requires(req)
            # not quite sure about what do to here in fact
            else:
                print("WARNING: fishy requirement in AbstractJob.requires")
                self.requires(list(requirement))

    def is_idle(self):
        """
        Returns:
          bool: ``True`` if the job has not been scheduled already, which
          in other words means that at least one of its requirements
          is not fulfilled.

        Implies `not is_scheduled()`, and so *a fortiori*
        `not is_running` and `not is_done()`.

        """
        return self._task is None

    def is_scheduled(self):
        """
        Returns:
          bool: ``True`` if the job has been scheduled.

        If True, it means that the job's requirements are met, and it has
        proceeded to the windowing system; equivalent to `not is_idle()`.
        """
        return self._task is not None

    def is_running(self):
        """
        Returns:
          bool: once a job starts, it tries to get a slot
          in the windowing sytem. This method returns ``True`` if the job
          has received the green light from the windowing system.
          Implies `is_scheduled()`.
        """
        return self._running

    def is_done(self):
        """
        Returns:
          bool: ``True`` if the job has completed.

        If this method returns ``True``, it implies that
        `is_scheduled()` and `is_running()`
        would also return ``True`` at that time.
        """
        return self._task is not None \
            and self._task._state == asyncio.futures._FINISHED

    def raised_exception(self):
        """
        Returns:
          an exception if the job has completed by raising an exception,
          and None otherwise.
        """
        return self._task is not None and self._task._exception

    def is_critical(self):
        """
        Returns:
          bool: whether this job is a critical job or not.
        """
        return self.critical

    def result(self):
        """
        Returns:
          When this job is completed and has not raised an exception, this
          method lets you retrieve the job's result. i.e. the value returned
          by its `co_run()` method.
        """
        if not self.is_done():
            raise ValueError("job not finished")
        return self._task._result

    async def co_run(self):
        """
        Abstract virtual - needs to be implemented
        """
        print("AbstractJob.co_run() needs to be implemented on class {}"
              .format(self.__class__.__name__))

    async def co_shutdown(self):
        """
        Abstract virtual - needs to be implemented
        """
        print("AbstractJob.co_shutdown() needs to be implemented on class {}"
              .format(self.__class__.__name__))

    def standalone_run(self):
        """
        A convenience helper that just runs this one job on its own.

        Mostly useful for debugging the internals of that job,
        e.g. for checking for gross mistakes and other exceptions.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.co_run())

    def details(self):
        """
        An optional method to implement on concrete job classes; if it
        returns a non None value, these additional details about that job
        will get printed by
        :meth:`asynciojobs.purescheduler.PureScheduler.list()` and
        :meth:`asynciojobs.purescheduler.PureScheduler.debrief()`
        when called with `details=True`.
        """
        pass


class Job(AbstractJob):

    """
    The simplest concrete job class, for building an instance of AbstractJob
    from of a python coroutine.

    Parameters:

      corun: a coroutine to be evaluated when the job runs
      coshutdown: an optional coroutine to be evaluated when the scheduler
        is done running
      scheduler: passed to :class:`AbstractJob`
      required: passed to :class:`AbstractJob`
      label: passed to :class:`AbstractJob`

    Example:
      To create a job that prints a message and waits for a fixed delay::

        async def aprint(message, delay):
            print(message)
            await asyncio.sleep(delay)

        j = Job(aprint("Welcome - idling for 3 seconds", 3))
    """

    def __init__(self, corun, *args, coshutdown=None, **kwds):
        self.corun = corun
        self.coshutdown = coshutdown
        super().__init__(*args, **kwds)

    def text_label(self):
        """
        Implementation of the method expected by :class:`AbstractJob`
        """
        try:
            return "Job[{name} (...)]".format(name=self.corun.__name__)
        except Exception:                                # pylint:disable=w0703
            return "Job instance"

    async def co_run(self):
        """
        Implementation of the method expected by :class:`AbstractJob`
        """
        result = await self.corun
        return result

    async def co_shutdown(self):
        """
        Implementation of the method expected by :class:`AbstractJob`,
        or more exactly by :meth:`asynciojobs.purescheduler.PureScheduler.list`
        """
        if self.coshutdown:
            result = await self.coshutdown
            return result
